import os
import base64
import magic

from typing import List, Dict
from .exception import SuprsendConfigError
from .workflow import WorkflowTrigger
from .request_log import set_logging


class Suprsend:
    """
    Entry point of SDK. Follow below Steps:
    1. create instance
        supr_client = Suprsend("__env_key__", "__env_secret__")
    2. Trigger required method
        workflow_body = {...}
        response = supr_client.trigger_workflow(workflow_body)
    """
    def __init__(self, env_key: str, env_secret: str, base_url: str = None, debug: bool = False, **kwargs):
        self.env_key = env_key
        self.env_secret = env_secret
        self.base_url = self.__get_url(base_url, kwargs)
        # include cryptographic signature
        self.auth_enabled = (kwargs.get('auth_enabled') is not False)
        self.include_signature_param = (kwargs.get('include_signature_param') is not False)
        # ---
        self.__validate()
        # --- set logging level for http request
        req_log_level = 1 if debug else 0
        set_logging(req_log_level)

    DEFAULT_URL = "https://hub.suprsend.com/"
    DEFAULT_UAT_URL = "https://collector-staging.suprsend.workers.dev/"

    def __get_url(self, base_url, kwargs):
        # ---- strip
        if base_url:
            base_url = base_url.strip()
        # ---- if url not passed, set url based on server env
        if not base_url:
            if kwargs.get("is_uat", False):
                base_url = self.DEFAULT_UAT_URL
            else:
                base_url = self.DEFAULT_URL
        # ---- check url ends with /
        base_url = base_url.strip()
        if base_url[len(base_url) - 1] != "/":
            base_url = base_url + "/"
        return base_url

    def __validate(self):
        if not self.env_key:
            raise SuprsendConfigError("Missing env_key")
        if not self.env_secret:
            raise SuprsendConfigError("Missing env_secret")
        if not self.base_url:
            raise SuprsendConfigError("Missing base_url")

    def get_attachment_json_for_file(self, file_path: str) -> Dict:
        # Ensure that path is expanded and absolute
        abs_path = os.path.abspath(os.path.expanduser(file_path))
        found = os.path.isfile(abs_path)
        if not found:
            return {
                "success": False,
                "error": "Not a file: {}".format(abs_path),
                "attachment": {}
            }
        # Get attachment json
        try:
            with open(abs_path, "rb") as f:
                file_name = os.path.basename(abs_path)
                mime_type = magic.from_file(abs_path, mime=True)
                # base64 encoded string
                b64encoded = base64.b64encode(f.read())
                b64data = b64encoded.decode()
                return {
                    "success": True,
                    "error": None,
                    "attachment": {
                        "filename": file_name,
                        "contentType": mime_type,
                        "data": b64data,
                    }
                }
        except FileNotFoundError as fnfe:
            return {
                "success": False,
                "error": str(fnfe),
                "attachment": {}
            }

    def trigger_workflow(self, data: Dict) -> Dict:
        """
        :param data:
        :return: {
            "success": true/false,
            "status": 202/401/500,
            "message": "message",
        }
        :except:
            - SuprsendConfigError (if proper value for env_key and env_secret not set
            - SuprsendValidationError (if post-data is invalid.)
        """
        wt = WorkflowTrigger(self, data)
        wt.validate_data()
        return wt.execute_workflow()



from datetime import datetime, timezone
import requests
import jsonschema
from urllib.parse import quote_plus
import json
from .exception import SuprsendValidationError, SuprsendInvalidSchema
from .request_schema import _get_schema
from .signature import get_request_signature


# In TZ Format: "%a, %d %b %Y %H:%M:%S %Z"
HEADER_DATE_FMT = "%a, %d %b %Y %H:%M:%S GMT"


class WorkflowTrigger:
    def __init__(self, config, data):
        self.config = config
        self.data = data
        self.url = self.__get_url()

    def __get_url(self):
        url_template = "{}{}/trigger/"
        if self.config.include_signature_param:
            if self.config.auth_enabled:
                url_template = url_template + "?verify=true"
            else:
                url_template = url_template + "?verify=false"
        url_formatted = url_template.format(self.config.base_url, self.config.env_key)
        # ---
        # self.url = quote_plus(url_formatted)
        return url_formatted

    def __get_headers(self):
        return {
            "Content-Type": "application/json",
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def execute_workflow(self):
        headers = self.__get_headers()
        # Based on whether signature is required or not, add Authorization header
        if self.config.auth_enabled:
            # Signature and Authorization-header
            content_txt, sig = get_request_signature(self.url, 'POST', self.data, headers, self.config.env_secret)
            headers["Authorization"] = "{}:{}".format(self.config.env_key, sig)
        else:
            content_txt = json.dumps(self.data)
        # -----
        resp = requests.post(self.url,
                             data=content_txt,
                             headers=headers)

        success = resp.status_code // 100 == 2
        return {
            "success": success,
            "status": resp.status_code,
            "message": resp.text,
        }

    def validate_data(self):
        schema = _get_schema('workflow')
        try:
            # jsonschema.validate(instance, schema, cls=None, *args, **kwargs)
            jsonschema.validate(self.data, schema)
        except jsonschema.exceptions.SchemaError as se:
            raise SuprsendInvalidSchema(se.message)
        except jsonschema.exceptions.ValidationError as ve:
            raise SuprsendValidationError(ve.message)
        return self.data



class SuprsendError(Exception):
    """
    A base class for all Suprsend exceptions.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.message = kwargs.get('message')
        if self.message is None and self.args and self.args[0]:
            self.message = self.args[0]
        elif self.message is None:
            self.message = None
        self.status_code = kwargs.get('status_code', 500)


class SuprsendConfigError(SuprsendError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SuprsendAuthenticationError(SuprsendError):
    """
    Invalid auth
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status_code = 401
        self.error_type = "SuprsendAuthenticationError"
        if self.message is None:
            self.message = "Authentication failed"


class SuprsendAuthorizationError(SuprsendError):
    """
    client does not have authorization to access API.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status_code = 403
        self.error_type = "SuprsendAuthorizationError"
        if self.message is None:
            self.message = "Authorization failed"


class SuprsendMissingSchema(SuprsendError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status_code = 500
        self.error_type = "SuprsendMissingSchema"
        if self.message is None:
            self.message = "Missing JSON schema"


class SuprsendInvalidSchema(SuprsendError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status_code = 500
        self.error_type = "SuprsendInvalidSchema"
        if self.message is None:
            self.message = "Invalid JSON schema"


class SuprsendValidationError(SuprsendError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status_code = 400
        self.error_type = "SuprsendValidationError"
        if self.message is None:
            self.message = "validation error"
